# Globals #

import os.path
import sys
import pandas as pd
from pprint import pprint

sys.path.insert(0, "modules")
# Project modules
from model import Model
import preprocessing as pp
import analysis
import scraper

def main():
    print("Fetching data...")

    price_data = scraper.fetch_data(os.path.dirname(os.getcwd()) + "/data/price_data.csv")
    blockchain_data = scraper.fetch_data(os.path.dirname(os.getcwd()) + "/data/blockchain_data.csv")

    print("Preprocessing...")

    data = (
        price_data.pipe(pp.calculate_indicators)
        .pipe(pp.merge_datasets, other_sets=[blockchain_data])
        .pipe(pp.binarize_labels)
        .pipe(pp.fix_null_vals)
        .pipe(pp.add_lag_variables, lag=3)
        .pipe(pp.power_transform)
    )
    x_train, x_test, y_train, y_test = pp.split(data, test_size=.2, balanced=True)

    # print("Analyzing features...")
    #
    # print(data.describe())
    # analysis.plot_corr_matrix(data)

    print("Fitting models...")

    log_reg = Model(
        estimator="LogisticRegression",
        train_set=(x_train, y_train),
        test_set=(x_test, y_test),
        optimize=True
    )
    rand_forest = Model(
        estimator="RandomForest",
        train_set=(x_train, y_train),
        test_set=(x_test, y_test),
        optimize=True
    )
    grad_boost = Model(
        estimator="GradientBoosting",
        train_set=(x_train, y_train),
        test_set=(x_test, y_test),
        optimize=True
    )

    print("Evaluating models...")

    print("\tLogistic Regression Estimator")
    log_reg.plot_cnf_matrix()
    log_reg.cross_validate(method="Holdout")
    # log_reg.cross_validate(
    #     method="RollingWindow",
    #     data=data,
    #     window_size=.9,
    #     test_size=.1
    # )

    print("\tRandom Forest Classifier")
    rand_forest.plot_cnf_matrix()
    rand_forest.cross_validate(method="Holdout")
    # rand_forest.cross_validate(
    #     method="RollingWindow",
    #     data=data,
    #     window_size=.9,
    #     test_size=.1
    # )

    print("\tGradient Boosting Classifier")
    grad_boost.plot_cnf_matrix()
    grad_boost.cross_validate(method="Holdout")
    # grad_boost.cross_validate(
    #     method="RollingWindow",
    #     data=data,
    #     window_size=.9,
    #     test_size=.1
    # )

if __name__ == '__main__':
    main()
